
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "cell.h"
#include "rif_rgb.h"
#include "sif.h"
#include "sif_rgb.h"

extern void	Error();

#define STR_LEN	1024
#define FPUTS	(void)fputs
#define SOC	sizeof(Cell)

typedef struct List {
		char		*path;
		int		h1,v1,h2,v2,x0,y0;
		struct List	*next;
	} List;

int	main(argc,argv)
int	argc;
char	**argv;
{
	int	Nx,Ny,x1,y1,x2,y2,R,G,B,tmp,h1,v1,h2,v2,x0,y0,Nh,Nv,y,x,v,h;
	char	line[STR_LEN],path[STR_LEN];
	List	*list,*next,**ptr=(&next);
	Cell	**TR,**TG,**TB,**SR,**SG,**SB;
	double	ratio;

	switch(argc) {
	    case 5 :
		break;
	    case 7 :
		if ((Nx=atoi(argv[1]))<=0 ||
		    (Ny=atoi(argv[2]))<=0) Error("bad image size.");

		x1=y1=0; x2=Nx-1; y2=Ny-1; break;
	    case 9 :
		if ((x1=atoi(argv[1]))>
		    (x2=atoi(argv[3])) ||
		    (y1=atoi(argv[2]))>
		    (y2=atoi(argv[4]))) Error("bad image area.");

		Nx=x2-x1+1; Ny=y2-y1+1; break;
	    default:
		Error("usage : paste_RGB R G B TIFF\n"
		      "        paste_RGB Nx Ny R G B TIFF\n"
		      "        paste_RGB x1 y1 x2 y2 R G B TIFF");
	}
	if ((R=atoi(argv[argc-4]))<0 || R>255 ||
	    (G=atoi(argv[argc-3]))<0 || G>255 ||
	    (B=atoi(argv[argc-2]))<0 || B>255)
	    Error("bad R, G and/or B value for space filling.");

#define ERROR_LINE(msg)	FPUTS(msg " :\n",stderr); Error(line)

	while (fgets(line,STR_LEN,stdin)!=NULL) {
	    if (line[tmp=strlen(line)-1]!='\n') {
		ERROR_LINE("too long line");
	    }
	    line[tmp]='\0';

	    switch(tmp=sscanf(line,
			      "%s %d %d %d %d %d %d",
			      path,&h1,&v1,&h2,&v2,&x0,&y0)) {
		case EOF: continue;
		case 1  : h1=v1=x0=y0=0; break;
		case 3  : x0=h1; y0=v1; h1=v1=0; break;
		case 5  : x0=y0=0; break;
		case 7  : break;
		default : ERROR_LINE("bad number of parameters");
	    }
	    ReadImageFile_RGB(path,&Nh,&Nv,NULL,NULL,NULL,NULL,NULL);

	    if (tmp==1 || tmp==3) {
		h2=Nh-1; v2=Nv-1;
	    }
	    else
		if (h1<0 || h2>=Nh || v1<0 || v2>=Nv) {
		    ERROR_LINE("bad image area");
		}

	    if ((*ptr=list=(List *)malloc(sizeof(List)))==NULL ||
		(list->path=strdup(path))==NULL)
		Error("list memory allocation error.");

	    list->h1=h1; list->h2=h2; list->x0=x0;
	    list->v1=v1; list->v2=v2; list->y0=y0; ptr=(&(list->next));
	}
	*ptr=NULL;

	if (argc==5) {
	    if (next==NULL) Error("no image.");

	    list=next; x2=(x1=list->x0)+list->h2-list->h1;
		       y2=(y1=list->y0)+list->v2-list->v1;

	    while ((list=list->next)!=NULL) {
		if ((x0=list->x0)<x1) x1=x0;
		if ((y0=list->y0)<y1) y1=y0;

		if ((tmp=x0+list->h2-list->h1)>x2) x2=tmp;
		if ((tmp=y0+list->v2-list->v1)>y2) y2=tmp;
	    }
	    Nx=x2-x1+1; Ny=y2-y1+1;
	}
	if ((TR=(Cell **)malloc(sizeof(Cell *)*(size_t)Ny))==NULL ||
	    (*TR=(Cell *)malloc(SOC*(size_t)Ny*(size_t)Nx))==NULL ||
	    (TG=(Cell **)malloc(sizeof(Cell *)*(size_t)Ny))==NULL ||
	    (*TG=(Cell *)malloc(SOC*(size_t)Ny*(size_t)Nx))==NULL ||
	    (TB=(Cell **)malloc(sizeof(Cell *)*(size_t)Ny))==NULL ||
	    (*TB=(Cell *)malloc(SOC*(size_t)Ny*(size_t)Nx))==NULL)
	    Error("cell memory allocation error.");

	for (y=0; y<Ny; y++) {
	    if (y>0) {
		TR[y]=TR[y-1]+Nx; TG[y]=TG[y-1]+Nx; TB[y]=TB[y-1]+Nx;
	    }
	    for (x=0; x<Nx; x++) {
		TR[y][x]=(Cell)R; TG[y][x]=(Cell)G; TB[y][x]=(Cell)B;
	    }
	}
	while ((list=next)!=NULL) {
	    if ((x0=list->x0)<=x2 && (x=x0+(h2=list->h2)-(h1=list->h1))>=x1 &&
		(y0=list->y0)<=y2 && (y=y0+(v2=list->v2)-(v1=list->v1))>=y1) {
		ReadImageFile_RGB(list->path,&Nh,&Nv,&tmp,&SR,&SG,&SB,NULL);

		if (Nh<=h2 || Nv<=v2) {
		    FPUTS(list->path,stderr); Error(" : bad image size.");
		}
		if (x>x2) h2=h1+x2-x0;
		if (y>y2) v2=v1+y2-y0;

		if (x0>=x1)
		    x0-=x1;
		else {
		    h1+=(x1-x0); x0=0;
		}
		if (y0>=y1)
		    y0-=y1;
		else {
		    v1+=(y1-y0); y0=0;
		}
		ratio=255.0/(double)((1<<tmp)-1);

		for (y=y0, v=v1; v<=v2; v++, y++)
		for (x=x0, h=h1; h<=h2; h++, x++) {
		    TR[y][x]=(Cell)(ratio*(double)SR[v][h]+0.5);
		    TG[y][x]=(Cell)(ratio*(double)SG[v][h]+0.5);
		    TB[y][x]=(Cell)(ratio*(double)SB[v][h]+0.5);
		}
		for (v=0; v<Nv; v++) {
		    free(SR[v]); free(SG[v]); free(SB[v]);
		}
		free(SR); free(SG); free(SB);
	    }
	    next=list->next; free(list);
	}
	for (y=0; y<Ny; y++)
	for (x=0; x<Nx; x++)
	    if (TR[y][x]!=TG[y][x] || TR[y][x]!=TB[y][x]) goto Break;

Break:;
	if (x==Nx && y==Ny)
	    StoreImageFile(argv[argc-1],Nx,Ny,8,TR,NULL);
	else
	    StoreImageFile_RGB(argv[argc-1],Nx,Ny,8,TR,TG,TB,NULL);

	return 0;
}
