
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "cell.h"
#include "csi.h"
#include "sif.h"

extern void	Error(),ReadSliceImage();

#define LEN	1024
#define SOC	sizeof(Cell)

typedef struct List {
		char		**paths;
		int		Nh,Nv,h1,v1,d1,h2,v2,d2,x0,y0,z0;
		struct List	*next;
	} List;

int	main(argc,argv)
int	argc;
char	**argv;
{
	int	Nx,Ny,Nz,x1,y1,z1,x2,y2,z2,PV,tmp,
		h1,v1,d1,h2,v2,d2,x0,y0,z0,Nh,Nv,Nd,v,y,z,x,d,h,pv,nh=0,nv=0;
	char	line[LEN],dir[LEN],name[LEN],**paths,form[LEN],path[LEN];
	List	*list,*next,**ptr=(&next);
	Cell	**S,**T;

	switch(argc) {
	    case 3 :
		break;
	    case 6 :
		if ((Nx=atoi(argv[1]))<=0 ||
		    (Ny=atoi(argv[2]))<=0 ||
		    (Nz=atoi(argv[3]))<=0) Error("bad image size.");

		x1=y1=z1=0; x2=Nx-1; y2=Ny-1; z2=Nz-1; break;
	    case 9 :
		if ((x1=atoi(argv[1]))>
		    (x2=atoi(argv[4])) ||
		    (y1=atoi(argv[2]))>
		    (y2=atoi(argv[5])) ||
		    (z1=atoi(argv[3]))>
		    (z2=atoi(argv[6]))) Error("bad image area.");

		Nx=x2-x1+1; Ny=y2-y1+1; Nz=z2-z1+1; break;
	    default:
		Error("usage : si_paste PV directory\n"
		      "        si_paste Nx Ny Nz PV directory\n"
		      "        si_paste x1 y1 z1 x2 y2 z2 PV directory");
	}
	if ((PV=atoi(argv[argc-2]))<0 || PV>=1<<8*SOC)
	    Error("bad pixel value for space filling.");

#define ERROR_LINE(msg)	(void)fputs(msg " :\n",stderr); Error(line)

	while (fgets(line,LEN,stdin)!=NULL) {
	    if (line[tmp=strlen(line)-1]!='\n') {
		ERROR_LINE("too long line");
	    }
	    line[tmp]='\0';

	    switch(tmp=sscanf(line,
			      "%s %s %d %d %d %d %d %d %d %d %d",
			      dir,name,&h1,&v1,&d1,&h2,&v2,&d2,&x0,&y0,&z0)) {
		case EOF: continue;
		case  2 : h1=v1=d1=x0=y0=z0=0; break;
		case  5 : x0=h1; y0=v1; z0=d1; h1=v1=d1=0; break;
		case  8 : x0=y0=z0=0; break;
		case 11 : break;
		default : ERROR_LINE("bad number of parameters");
	    }
	    paths=CheckSliceImages(dir,name,&Nh,&Nv,&Nd,NULL);

	    if (tmp==2 || tmp==5) {
		h2=Nh-1; v2=Nv-1; d2=Nd-1;
	    }
	    else
		if (h1<0 || h2>=Nh || v1<0 || v2>=Nv || d1<0 || d2>=Nd) {
		    ERROR_LINE("bad image area");
		}

	    if ((*ptr=list=(List *)malloc(sizeof(List)))==NULL)
		Error("no allocatable memory for list.");

	    list->paths=paths;
	    if ((list->Nh=Nh)>nh) nh=Nh;
	    if ((list->Nv=Nv)>nv) nv=Nv;
	    list->h1=h1; list->v1=v1; list->d1=d1;
	    list->h2=h2; list->v2=v2; list->d2=d2;
	    list->x0=x0; list->y0=y0; list->z0=z0; ptr=(&(list->next));
	}
	*ptr=NULL;

	if (argc==3) {
	    if (next==NULL) Error("no image.");

	    list=next; x2=(x1=list->x0)+list->h2-list->h1;
		       y2=(y1=list->y0)+list->v2-list->v1;
		       z2=(z1=list->z0)+list->d2-list->d1;

	    while ((list=list->next)!=NULL) {
		if ((x0=list->x0)<x1) x1=x0;
		if ((y0=list->y0)<y1) y1=y0;
		if ((z0=list->z0)<z1) z1=z0;

		if ((tmp=x0+list->h2-list->h1)>x2) x2=tmp;
		if ((tmp=y0+list->v2-list->v1)>y2) y2=tmp;
		if ((tmp=z0+list->d2-list->d1)>z2) z2=tmp;
	    }
	    Nx=x2-x1+1; Ny=y2-y1+1; Nz=z2-z1+1;
	}
	if ((S=(Cell **)malloc(sizeof(Cell *)*(size_t)nv))==NULL ||
	    (*S=(Cell *)malloc(SOC*(size_t)nv*(size_t)nh))==NULL ||
	    (T=(Cell **)malloc(sizeof(Cell *)*(size_t)Ny))==NULL ||
	    (*T=(Cell *)malloc(SOC*(size_t)Ny*(size_t)Nx))==NULL)
	    Error("no allocatable memory for cell.");

	for (v=1; v<nv; v++) S[v]=S[v-1]+nh;
	for (y=1; y<Ny; y++) T[y]=T[y-1]+Nx;

	tmp=1; for (z=10; z<Nz; z*=10) ++tmp;
	(void)sprintf(form,"%s/%%0%dd.tif",argv[argc-1],tmp);

	for (z=0; z<Nz; z++) {
	    for (y=0; y<Ny; y++)
	    for (x=0; x<Nx; x++) T[y][x]=(Cell)PV;

	    for (list=next; list!=NULL; list=list->next)
		if ((x0=list->x0)<=x2 &&
		    (x=x0+(h2=list->h2)-(h1=list->h1))>=x1 &&
		    (y0=list->y0)<=y2 &&
		    (y=y0+(v2=list->v2)-(v1=list->v1))>=y1 &&
		    (d=z+z1-list->z0)>=0 && (d+=list->d1)<=list->d2) {
		    ReadSliceImage(list->paths[d],list->Nh,list->Nv,S);

		    if (x>x2) h2=h1+x2-x0;
		    if (y>y2) v2=v1+y2-y0;

		    if (x0>=x1)
			x0-=x1;
		    else {
			h1+=(x1-x0); x0=0;
		    }
		    if (y0>=y1)
			y0-=y1;
		    else {
			v1+=(y1-y0); y0=0;
		    }
		    for (y=y0, v=v1; v<=v2; v++, y++)
		    for (x=x0, h=h1; h<=h2; h++, x++) T[y][x]=S[v][h];
		}

	    pv=0;
	    for (y=0; y<Ny; y++)
	    for (x=0; x<Nx; x++) if (T[y][x]>pv) pv=T[y][x];

	    (void)sprintf(path,form,z);
	    for (tmp=1; 1<<tmp<=pv; tmp<<=1) ;
	    StoreImageFile(path,Nx,Ny,tmp,T,NULL);
	}
	return 0;
}
